/*-----------------------------------------------------------------------
QMBTagger - Qt Based MusicBrainz MP3/OGG/FLAC Tagger
Copyright (C) 2003,2004 Ewen Cheslack-Postava

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-----------------------------------------------------------------------*/

#include <config.h>

#include "qmbtagbox.h"
#include "qmbfileview.h"
#include "qmbfile.h"
#include "qmbfileviewfileitem.h"
#include "qmbfileviewresultitem.h"
#include "qmbtagger.h"

#include "qmbvorbisfile.h"
#include "qmbmp3file.h"
#include "qmbflacfile.h"

#include <qlistview.h>
#include <qstring.h>
#include <qstringlist.h>
#include <qapplication.h>
#include <qmessagebox.h>
#include <qprogressdialog.h>
#include <qsettings.h>
#include <qaction.h>
#include <qdragobject.h>
#include <qdir.h>
#include <qvbox.h>
#include <qlistview.h>
#include <qlistbox.h>
#include <qpopupmenu.h>
#include <musicbrainz/musicbrainz.h>
#include <musicbrainz/mb_c.h>
#include <algorithm>

QMBFileView::QMBFileView(QWidget* parent, const char* name)
	:QSplitter(parent, name), format("{ARTIST} - {TRACK}.{EXT}"), newdir(""), crossfs(true), submit(), authenticated(false), username(""), password(""), tagbox(NULL), contextItem(NULL)
{
	setAcceptDrops(TRUE);
	
	songs = new QListBox(this);
	
	details = new QSplitter(QSplitter::Vertical, this);
	
	metalist = new QListView(details);
	metalist->addColumn("Variable");
	metalist->addColumn("Value");
	
	results = new QListView(details);
	results->addColumn("Name");
	results->addColumn("Artist");
	results->addColumn("Album");
	results->addColumn("Title");
	results->addColumn("TrackNum");
	results->addColumn("Length");
	results->addColumn("Relevance");
	results->setSelectionMode(QListView::Multi);
	//results->setRootIsDecorated(true);

	mb = new MusicBrainz();

	//set the default server info
	mb->SetServer("musicbrainz.org", 80);

	//set it to not use UTF8
	mb->UseUTF8(FALSE);

	connect(this, SIGNAL(SaveTRMs()), &submit, SLOT(Save()));
	connect(this, SIGNAL(LoadTRMs()), &submit, SLOT(Load()));

	connect(songs, SIGNAL(contextMenuRequested(QListBoxItem*, const QPoint&)), this, SLOT(showContextMenu(QListBoxItem*, const QPoint &)));
	connect(songs, SIGNAL(doubleClicked(QListBoxItem*)), this, SLOT(ManuallyTag(QListBoxItem*)));
	connect(songs, SIGNAL(clicked ( QListBoxItem *)), this, SLOT(songClicked(QListBoxItem*)));
	
	connect(results, SIGNAL(clicked ( QListViewItem *)), this, SLOT(resultClicked(QListViewItem*)));
	
	tagbox = new QMBTagBox();
	tagbox->setCaption("Custom Tags");
	
	connect(tagbox, SIGNAL(ChangedMetadata()), this, SLOT(AppliedTag()));
	
	contextMenu = new QPopupMenu();
	
	QAction* editTagData = new QAction("Manually Edit Tag Data", "&Edit Tags", CTRL+Key_E, this, "edit");
	editTagData->addTo(contextMenu);
	connect(editTagData, SIGNAL(activated()), this, SLOT(ManuallyTag()));
	
	QAction* removeSongFromList = new QAction("Remove Song From List", "Remove Song", Key_Delete, this, "remove");
	removeSongFromList->addTo(contextMenu);
	connect(removeSongFromList, SIGNAL(activated()), this, SLOT(RemoveSong()));
	
	return;
}

QMBFileView::~QMBFileView()
{
	//delete the musicbrainz object
	if (mb != NULL)
		delete mb;

	//delete all the files we have created
	QMBFileViewFileItem* curfileitem = NULL;
	for (vector<QMBFileViewFileItem*>::iterator fileindex = listparents.begin(); fileindex != listparents.end(); fileindex++) {
		curfileitem = (*fileindex);
		if (curfileitem != NULL)
			delete curfileitem;
	}

}

int QMBFileView::NumSavedTRMs() {
	return submit.Count();
}

int QMBFileView::FileCount() {
	return listparents.size();
}

bool QMBFileView::Authenticate() {
	if (this->authenticated) return true;

	//no need to actually store the username and password, just authenticate and warn if something is wrong
	if (!this->username.isEmpty() && !this->password.isEmpty() && !mb->Authenticate((std::string)this->username.latin1(), (std::string)this->password.latin1())) {
		qWarning("QMBFileView::Authenticate(): Authentication failed.");
		QMessageBox::information(this, "Authentication error...", "Failed to authenticate with the server.  Check your username.", QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);
		return false;
	}

	this->authenticated = true;
	return true;
}

void QMBFileView::TransmitSettings(QSettings* newsettings) {
	this->submit.SetSettings(newsettings);
}

void QMBFileView::SetFilenames(const QStringList &filelist) {

	int fileindex, filecount = filenames.count(), newfilecount = filelist.count();
	int totalfilecount = filecount + newfilecount;

	if (newfilecount < 1) {
		return;
	}
	
	//add the new items
	for( fileindex = filecount; fileindex < totalfilecount; fileindex++) {
		filenames.append(filelist[fileindex - filecount]);
		QString curfilename = filenames[fileindex];

		QString shortfilename = curfilename.right( curfilename.length() - curfilename.findRev("/") - 1);

		//add the parent level item without any information
		QMBFileViewFileItem* toplevel = new QMBFileViewFileItem(songs, shortfilename);
		if (toplevel->CreateFile(curfilename) == NULL) {
			delete toplevel;
			continue;
		}
		toplevel->setSelectable(TRUE);
		//toplevel->setOpen(TRUE);
		
		//insert into the list dictionary
		listparents.push_back(toplevel);
	}
}

bool QMBFileView::AddDirectory(QString dirname, bool recursive) {
	int numfiles, index;
	QDir dir(dirname);
	bool result = false;
	
	if (recursive) {
		dir.setFilter(QDir::Dirs);
		QStringList subdirs = dir.entryList();
		numfiles = subdirs.count();
		for (index = 0; index < numfiles; index++) {
			if (subdirs[index] != "." && subdirs[index] != ".." && AddDirectory(dirname + subdirs[index] + "/", true))
				result = true;
		}
	}
	
	dir.setFilter(QDir::Files | QDir::Hidden);
	dir.setNameFilter(QMBTagger::GetFileTypeString());
	
	//this makes the directory be included in the file name
	QStringList files = dir.entryList();
	numfiles = files.count();
	for(index = 0; index < numfiles; index++) {
		files[index].prepend(dirname);
	}

	SetFilenames(files);
	if (numfiles > 0)
		result = true;
	
	return result;
}

void QMBFileView::QueryServer() {

	if (listparents.empty()) {
		QMessageBox::information(this, "No files to query...", "QMBFileView::QueryServer(): No filenames selected for tagging.");
		return;
	}
	else if (listparents.size() > 10) {
		//ask if the user really wants to query so many items
		int result = QMessageBox::warning(this, "Query?", "You have chosen to look up " + QString::number(listparents.size()) + " song.  This process can take a long time.  Continue?", "Query", "Don't Query", QString::null, 0, 1);
		if (result == 1) {
			return;
		}
	}

	int fileindex, filecount = listparents.size();

	//setup the progress dialog
	QProgressDialog progress("Querying server...", "Cancel", 3*filecount, this, "progress", TRUE);
	//make the progress dialog show no matter how long it takes
	progress.setMinimumDuration(0);
	progress.setCaption("Query progress...");
	progress.setProgress(0);
	
	for( fileindex = 0; fileindex < filecount; fileindex++) {

		QMBFileViewFileItem* toplevel = listparents[fileindex];
		QMBFile* curfile = toplevel->GetFile();

		if (toplevel->ChildCount() > 0) continue; //don't process files that already have children, we know they have already been queried

		QString curfilename = filenames[fileindex];

		QString shortfilename = curfilename.right( curfilename.length() - curfilename.findRev("/") - 1);

		trm = trm_New();
		
		//the following three steps are just using the
		//functionality of the QMBFile class
		if (curfile->metadata.GetTRMID().isEmpty()) { //only extract and create ids for files that don't have ids yet
			if (curfile->Extract(curfilename) < 0) {
				qWarning("QMBFileView::QueryServer(): The current file (%s) couldn't be extracted.", curfilename.latin1());
				continue;
			}

			emit ExtractedFileData("Extracted Data From: " + shortfilename);
			//increment the progress dialog
			progress.setProgress(3*fileindex + 1);
			qApp->processEvents();
			if (progress.wasCancelled()) break;

			int idCreationResult = curfile->CreateID(trm);
			trm_Delete(trm);
			
			if (idCreationResult < 0) {
				qWarning("QMBFileView::QueryServer(): The current file (%s) couldn't have an ID properly made.", curfilename.latin1());
				continue;
			}

			emit CreatedId("Created ID for: " + shortfilename);
			//increment the progress dialog
			progress.setProgress(3*fileindex + 2);
			qApp->processEvents();
			if (progress.wasCancelled()) break;
		}

		int numtracks = curfile->QueryServer(mb);
		emit QueriedServer("Queried Server for: " + shortfilename);
		qApp->processEvents();
		if (numtracks < 1) {
			qWarning("QMBFileView::QueryServer(): The current file (%s) did not successfully get results from the server.", curfilename.latin1());
			continue;
		}
			
		//now we just add all the results
		QMBFileViewResultItem* child = NULL;
		QString ext = curfile->GetFileExtension();
		for (int resultindex = 0; resultindex < numtracks; resultindex++) {
			QString newfilename = format;
			curfile->queryresults[resultindex].ReplaceFilenameWildcards(newfilename, ext);
			child = new QMBFileViewResultItem(results, newfilename, curfile->queryresults[resultindex].GetArtistName(), curfile->queryresults[resultindex].GetAlbumName(), curfile->queryresults[resultindex].GetTrackName(), curfile->queryresults[resultindex].GetTrackNumber(), curfile->queryresults[resultindex].GetFormattedDuration(), QString::number(curfile->queryresults[resultindex].GetRelevance()));
			child->SetIndex(resultindex);
			toplevel->AddChild(child);
		}

		toplevel->HideChildren(results);
		
		emit AddedResultsToList("Added results for: " + shortfilename);
		emit AddedResultsToList();

		//increase the progress
		progress.setProgress(3*(fileindex + 1));

		qApp->processEvents();

		if (progress.wasCancelled()) break;
	}
	
	//if a song is selected, display its results
	songClicked(songs->selectedItem());
	
	emit FinishedQuerying("Finished Querying.");

	return;
}

void QMBFileView::OutputSelected() {

	//ask if the user really wants to tag all these items
	int result = QMessageBox::warning(this, "Tag Songs?", "Are you sure you want to tag all selected songs?", "Tag", "Don't Tag", QString::null, 0, 1);
	if (result == 1) {
		return;
	}

	//setup the progress dialog
	QProgressDialog progress("Saving files...", "Cancel", songs->count(), this, "progress", TRUE);
	//make the progress dialog show no matter how long it takes
	progress.setMinimumDuration(0);
	progress.setCaption("Saving progress...");
	progress.setProgress(0);
	
	for(unsigned int fileindex = 0; fileindex < listparents.size(); fileindex++) {
		//if this is a parent item then we check for manually tagged data
		QMBFileViewFileItem* currentfileitem = listparents[fileindex];
		QMBFile* currentfile = currentfileitem->GetFile();
			
		if (currentfile->wasManuallyTagged() == true) {
			//was manually tagged, output it and skip over the children
			if (currentfile->Output(-1, format, newdir, crossfs) != 0) {
				//we errored out somewhere in the output functions
				qWarning("QMBFileView::OutputSelected(): Error outputting file (%s).", currentfile->fullfilename.latin1());
				continue;
			} 
			
			//get rid of all the results
			currentfileitem->ClearChildren(results);
			
			//now remove the parent item
			songs->takeItem(currentfileitem);

			emit OutputtedFile("Outputted new file for: " + currentfile->fullfilename);
		}
		else {
			//find out which item is selected
			QMBFileViewResultItem* selectedresult = currentfileitem->GetSelectedChild();

			if (selectedresult != NULL) {
				//now actually do the outputting
				if (currentfile->Output(selectedresult->GetIndex(), format, newdir, crossfs) != 0) {
					//we errored out somewhere in the output functions
					qWarning("QMBFileView::OutputSelected(): Error outputting file (%s).", currentfile->fullfilename.latin1());
					continue;
				}
			
				//add the item to submit to the server
				submit.Add(currentfile->metadata.GetTrackID(), currentfile->metadata.GetTRMID());

				//get rid of all the results
				currentfileitem->ClearChildren(results);
			
				//now remove the parent item
				songs->takeItem(currentfileitem);

				emit OutputtedFile("Outputted new file for: " + currentfile->fullfilename);
			}
		}
		//increase the progress
		progress.setProgress(progress.progress() + 1);
		qApp->processEvents();
	}

	emit FinishedOutputting("Finished writing files.");

	//nothing left to do, if we are here we must have succeeded
	return;
}

void QMBFileView::setOptions(QString server, int port, QString proxy, int pport, QString filenameformat, QString outputdir, bool copycrossfs, QString user, QString pass) {
	//store the filename format so it can be passed to the output function later
	this->format = filenameformat;

	//the serverurl and serverport are set immediately in the MusicBrainz object
	if (this->mb != NULL) {
		if (!server.isEmpty())
			mb->SetServer(server.latin1(), port);
		if (!proxy.isEmpty()) {
			mb->SetProxy(proxy.latin1(), (short)pport);
			trm_SetProxy(trm, (char*)proxy.latin1(), (short)pport);
		}
	}

	//store the output directory so it can be used in the output function
	this->newdir = outputdir;

	this->crossfs = copycrossfs;

	//store the username and password here
	if (this->username != user || this->password != pass) {
		this->username = user;
		this->password = pass;

		this->authenticated = false;

		this->Authenticate();
	}
}

void QMBFileView::SubmitTRMs() {

	this->Authenticate();

	if (submit.Submit(mb)) {
		emit SubmittedTRMs("Submitted TRMs");
	} else {
		QMessageBox::information(this, "Submission error...", "Couldn't submit TRMs.  Check your password.", QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);
	}
}

void QMBFileView::showContextMenu(QListBoxItem* item, const QPoint & pos){
	//only show the context menu if the item is a parent (i.e. an actual file, not a result)
	if (item != NULL) {
		contextItem = (QMBFileViewFileItem*)item;
		contextMenu->popup(pos);
	}
}

void QMBFileView::ManuallyTag() {
	
	tagbox->SetMetadataPtr(&(contextItem->GetFile()->metadata));
	tagbox->show();
			
	return;
}

void QMBFileView::ManuallyTag(QListBoxItem* item) {
	tagbox->SetMetadataPtr(&(((QMBFileViewFileItem*)item)->GetFile()->metadata));
	tagbox->show();
			
	return;
}

void QMBFileView::AppliedTag() {
	contextItem->GetFile()->setManuallyTagged(true);
	contextItem = NULL;
}

void QMBFileView::dragEnterEvent(QDragEnterEvent* event)
{
    event->accept(QUriDrag::canDecode(event));
}

void QMBFileView::dropEvent(QDropEvent * e) {
	QStringList filedirlist;
	QStringList filelist;
	QDir dir;
	
	if (QUriDrag::decodeLocalFiles(e, filedirlist) ) {
		for (unsigned int index = 0; index < filedirlist.count(); index++) {
			if (dir.exists(filedirlist[index]))
				AddDirectory(filedirlist[index] + "/", true);
			else
				filelist.push_back(filedirlist[index]);
		}
		SetFilenames(filelist);
    } 
}

void QMBFileView::resultClicked(QListViewItem* item) {
	if (item == NULL) return;
		
	//if a an item was actually clicked and it is now selected, make sure
	//none of its siblings are selected
	if (item != NULL && item->isSelected()) ;
		((QMBFileViewFileItem*)songs->selectedItem())->SetSingleSelection(results, item);
}

void QMBFileView::songClicked(QListBoxItem* item) {
	if (item == NULL) return;
		
	//set up the metadata
	metalist->clear();
	QMBFile* curfile = ((QMBFileViewFileItem*)item)->GetFile();
	new QListViewItem(metalist, "Track", curfile->metadata.GetTrackName());
	new QListViewItem(metalist, "Artist", curfile->metadata.GetArtistName());
	new QListViewItem(metalist, "Album", curfile->metadata.GetAlbumName());
	new QListViewItem(metalist, "Track #", curfile->metadata.GetTrackNumber());
	new QListViewItem(metalist, "Duration", curfile->metadata.GetFormattedDuration());
	new QListViewItem(metalist, "Track ID", curfile->metadata.GetTrackID());
	new QListViewItem(metalist, "Artist ID", curfile->metadata.GetArtistID());
	new QListViewItem(metalist, "Album ID", curfile->metadata.GetAlbumID());
	new QListViewItem(metalist, "TRM ID", curfile->metadata.GetTRMID());
	
	//set up the results
	for (unsigned index = 0; index < listparents.size(); index++) {
		if (listparents[index] == (QMBFileViewFileItem*)item)
			listparents[index]->ShowChildren(results);
		else
			listparents[index]->HideChildren(results);
	}
}

void QMBFileView::RemoveSong() {
	if (contextItem == NULL)
		contextItem = (QMBFileViewFileItem*)songs->selectedItem();

	vector<QMBFileViewFileItem*>::iterator item = std::find(listparents.begin(), listparents.end(), contextItem);
	if (item != listparents.end())
		listparents.erase(item);
	contextItem->HideChildren(results);
	metalist->clear();
	songs->takeItem((QListBoxItem*)contextItem);
	delete contextItem;
	contextItem = NULL;
}
