/*-----------------------------------------------------------------------
QMBTagger - Qt Based MusicBrainz MP3/OGG/FLAC Tagger
Copyright (C) 2003,2004 Ewen Cheslack-Postava

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-----------------------------------------------------------------------*/

#include <config.h>

#if USE_MP3 == 1

#include "qmbmp3file.h"
#include "qmbfile.h"
#include <stdio.h>
#include <string.h>
#include <qapplication.h>
#include <qstring.h>
#include <musicbrainz/musicbrainz.h>
#include "qmbmad.h"
#include <id3/misc_support.h>
#include <fstream>

QMBMP3File::QMBMP3File()
	: QMBFile()
{
	//we do nothing here.  we want to allow checking of return value,
	//so we only allow class to function through calling of Extract()
}

QMBMP3File::~QMBMP3File()
{
	if (buffer != 0) {
		delete [] buffer;
		bufferlen = 0;
	}
	if (queryresults != NULL) {
		delete [] queryresults;
	}
}

//get the data we want from the vorbis stream
//heavily borrowed code from vorbis doc on decoding vorbis stream
//returns:
//	0  indicates success
//	-1 indicates failed to open
//	-2 indicates invalid info
//	-3 indicates decoder failure
//	1 indicates success, but metadata extraction failed
int QMBMP3File::Extract(QString filename) {
	
	FILE* file = NULL;
 	int bitrate, durationmilli, samplesint;
	int ret = 0;

	//record the full filename so we will be able to use it for output
	fullfilename = filename;

	//set the filename in the metadata
	int lastslash = filename.findRev('/');
	metadata.SetFileName(filename.right(filename.length() - lastslash - 1));

	//get the file's info
	MusicBrainz mb;
	std::string strfilename = filename.latin1(); //FIXME: change this to use QString if possible
	if (mb.GetMP3Info(strfilename, durationmilli, bitrate, channels, samplesint)) {
		metadata.SetDuration(QString::number(durationmilli));
		songlength = durationmilli/1000.;
		channels++; //this is because GetMP3Info reports single channel as 0 and stereo as 1
		sampleRate = samplesint;
		bits_per_sample = 16; //we set this in the decoder, so there is no possibility of it changing
	} else {
		qWarning("QMBMP3File::Extract(): GetMP3Info failed on file (%s).", fullfilename.latin1());
		return -2;
	}

	//if the info is invalid, return without doing anything
	if (channels == 0 || sampleRate == 0) {
		qWarning("QMBMP3File::Extract(): Channels (%d) or SampleRate (%ld) not set properly for file (%s).", channels, sampleRate, fullfilename.latin1());
		return -2;
	}

	//get metadata
	if (GetMetadata(fullfilename) == -1) {
		//getting metadata failed - very strange! indicate with return val
		qWarning("QMBMP3File::Extract(): GetMetadata failed: File (%s) must be corrupted.", fullfilename.latin1());
		ret = 1;
	}

	//if the trmid was read in with the metadata, quit now
	if (!metadata.GetTRMID().isEmpty())
		return 0;
	
	//now for the actual decoding
	file = fopen(fullfilename.latin1(), "r");
	if (file == NULL) {
		qWarning("QMBMP3File::Extract(): Couldn't open file (%s) for reading.", fullfilename.latin1());
		return -1;
	}

	//figure out buffer length, we want 30 seconds of data,
	//get rate and channels from vorbis file, and
	//specify 16 bit (2 byte) samples
	//= channels*30 sec*rate*2bytes
	bufferlen = 30*channels*sampleRate*bits_per_sample/8;

	//create the actual buffer
	buffer = new char[bufferlen];

	//read the mp3
	QMBmad::ReadMPEG(file, buffer, bufferlen);

	//we're finished, just need to clean up
	fclose(file);

	return ret;
}

int QMBMP3File::GetMetadata(QString filename) {
	ID3_Tag filetag(filename.latin1());

	//get the album name
	metadata.SetAlbumName(ID3_GetAlbum(&filetag));
	if (metadata.GetAlbumName().isNull()) metadata.SetAlbumName(""); //can't be null, but can be empty

	//get the song title
	metadata.SetTrackName(ID3_GetTitle(&filetag));
	if (metadata.GetTrackName().isNull()) metadata.SetTrackName(""); //can't be null, but can be empty

	//get the artist name
	metadata.SetArtistName(ID3_GetArtist(&filetag));
	if (metadata.GetArtistName().isNull()) metadata.SetArtistName(""); //can't be null, but can be empty

	//get the track number
	metadata.SetTrackNumber(ID3_GetTrack(&filetag));
	if (metadata.GetTrackNumber().isNull()) metadata.SetTrackNumber(""); //can't be null, but can be empty
	
	//we have to extract the custom MusicBrainz tags manually
	ID3_Tag::Iterator* tagIter = filetag.CreateIterator();
	ID3_Frame* curFrame = NULL;
	while (NULL != (curFrame = tagIter->GetNext()))
	{
		ID3_FrameID fid = curFrame->GetID();
		if (fid == ID3FID_UNIQUEFILEID) {
			char owner[1024];
			uchar trackid[37];
			size_t length = curFrame->GetField(ID3FN_OWNER)->Get(owner, 1024);
			owner[length] = '\0';
			if (strncmp(owner, "http://musicbrainz.org", 26) == 0) {
				length = curFrame->GetField(ID3FN_DATA)->Get(trackid, 36);
				if (length == 36) {
					trackid[36] = '\0';
					metadata.SetTrackID((char*)trackid);
				}
			}
		}
		else if (fid == ID3FID_USERTEXT) {
			char desc[1024];
			char id[37];
			size_t length;
			length = curFrame->GetField(ID3FN_DESCRIPTION)->Get(desc, 1024);
			desc[length] = '\0';
			length = curFrame->GetField(ID3FN_TEXT)->Get(id, 36);
			if (length == 36) {
				id[36] = '\0';
				if (strncmp(desc, "MusicBrainz Artist Id", 21) == 0)					
					metadata.SetArtistID(id);
				else if (strncmp(desc, "MusicBrainz Album Id", 20) == 0)
					metadata.SetAlbumID(id);
				else if (strncmp(desc, "MusicBrainz TRM Id", 18) == 0)
					metadata.SetTRMID(id);
			}
		}
	}
	delete tagIter;
	
	filetag.Clear();

	return 0;
}

//returns:
// 0 on succes
// -1 on invalid choice
// -2 on general fileerror (a couple of possibilities here)
// -3 on success, but old file couldn't be deleted
int QMBMP3File::Output(int resultChoice, QString format, QString newdir, bool crossfs) {

	//check if they are giving a valid choice
	if (resultChoice >= numresults) {
		qWarning("QMBMP3File::Output(): Invalid result choice - chose: %d, avail: %d.", resultChoice, numresults);
		return -1;
	}

	if (resultChoice >= 0) { //if we didn't manually tag
		//fill in metadata with metadata from selection
		this->CopySelectedMetadata(resultChoice);
	}
	
	//first do id3 operations on the original file
	ID3_Tag filetag(fullfilename.latin1());

	//strip out all other tags
	ID3_StripAllTags(&filetag);

	//set the album name
	ID3_AddAlbum(&filetag, metadata.GetAlbumName().latin1());

	//set the song title
	ID3_AddTitle(&filetag, metadata.GetTrackName().latin1());

	//set the artist name
	ID3_AddArtist(&filetag, metadata.GetArtistName().latin1());

	//set the track number
	ID3_AddTrack(&filetag, metadata.GetTrackNumber().toInt());

	//set the trackID
	ID3_AddTrackID(&filetag, metadata.GetTrackID().latin1());
	
	//set the artistID
	ID3_AddUserText(&filetag, "MusicBrainz Artist Id", metadata.GetArtistID().latin1());
	
	//set the albumID
	ID3_AddUserText(&filetag, "MusicBrainz Album Id", metadata.GetAlbumID().latin1());
	
	//set the trmID
	ID3_AddUserText(&filetag, "MusicBrainz TRM Id", metadata.GetTRMID().latin1());
	
	//update the tags
	filetag.SetPadding(false);
	filetag.Update(ID3TT_ALL);

	//disassociate the tagging object from the file
	filetag.Clear();

	//get the directory and then set the output name, replacing placeholders as we go
	QString newfilename = format;
	ReplaceFilenameWildcards(newfilename);
	
	if (newdir.isEmpty()) {
		//no newdir specified, so just use the original
		QString dir = fullfilename.left( fullfilename.findRev("/") + 1 );
		newfilename.prepend(dir);
	} else {
		if (newdir.right(1) != "/") newdir.append("/");
		newfilename.prepend(newdir);
	}

	//try to create the directory for the file if it isn't there yet
	if (!CreateDirectory(newfilename)) {
		qWarning("QMBMP3File::Output(): Couldn't create or navigate to the specified location: %s", newfilename.latin1());
		return -2;
	}
	
	//now we just have to move the file
	int ret = rename(fullfilename.latin1(), newfilename.latin1());
	if (ret < 0) {
		if (errno == EXDEV) {
			//couldn't rename across filesystems, we have to do a full copy instead
			if (crossfs) {
				//this is pretty efficient, but can take a long time on large files!
				std::ifstream fin(fullfilename.latin1(), STD_IOS_BASE_IN | STD_IOS_BASE_BINARY);
				std::ofstream fout(newfilename.latin1(), STD_IOS_BASE_OUT | STD_IOS_BASE_BINARY | STD_IOS_BASE_TRUNC);
				if (fin.is_open() && fout.is_open())
					fout << fin.rdbuf();
				fin.close();
				fout.close();
				if (remove(fullfilename.latin1()) < 0) {
					qWarning("QMBMP3File::Output(): Couldn't remove old file (%s) after successfully writing new file.", fullfilename.latin1());
					return -3;
				}
			} else {
				qWarning("QMBMP3File::Output(): You chose not to copy across filesystems.  The metadata has been written to the file (%s), but it has not been renamed.", fullfilename.latin1());
			}
		} else {
			qWarning("QMBMP3File::Output(): Error moving file from %s to %s.", fullfilename.latin1(), newfilename.latin1());
			return -2;
		}
	}

	//if we got here we have succeeded
	return 0;
}

QString QMBMP3File::GetFileExtension() {
	return QString("mp3");
}

ID3_Frame* QMBMP3File::ID3_AddTrackID(ID3_Tag *tag, const char *text) {
	ID3_Frame* frame = NULL;
	if (NULL != tag && NULL != text)
	{
		frame = new ID3_Frame(ID3FID_UNIQUEFILEID);
		if (frame)
		{
			frame->GetField(ID3FN_OWNER)->Set("http://musicbrainz.org");
			frame->GetField(ID3FN_DATA)->Set((const uchar*)text, strlen(text));
			tag->AttachFrame(frame);
		}
	}

	return frame;
}

ID3_Frame* QMBMP3File::ID3_AddUserText(ID3_Tag *tag, const char* desc, const char *text) {
	ID3_Frame* frame = NULL;
	if (NULL != tag && NULL != text)
	{
		frame = new ID3_Frame(ID3FID_USERTEXT);
		if (frame)
		{
			frame->GetField(ID3FN_TEXTENC)->SetEncoding(ID3TE_ISO8859_1);
			frame->GetField(ID3FN_DESCRIPTION)->Set(desc);
			frame->GetField(ID3FN_TEXT)->Set(text);
			tag->AttachFrame(frame);
		}
	}

	return frame;
}

void QMBMP3File::ID3_StripAllTags(ID3_Tag *tag) {
	ID3_Tag::Iterator* iter = tag->CreateIterator();
	ID3_Frame* curFrame = NULL;
	while (NULL != (curFrame = iter->GetNext()))
	{
		tag->RemoveFrame(curFrame);
	}
	delete iter;
}

#endif /* USE_MP3 == 1 */
